/*
 * THIS IS A GENERATED FILE
 * SEE COLUMNS.LIST
 */
#ifndef TABLES_H
#define TABLES_H
#define CALL_CENTER                0
#define CATALOG_PAGE               1
#define CATALOG_RETURNS            2
#define CATALOG_SALES              3
#define CUSTOMER                   4
#define CUSTOMER_ADDRESS           5
#define CUSTOMER_DEMOGRAPHICS      6
#define DATET                      7
#define HOUSEHOLD_DEMOGRAPHICS     8
#define INCOME_BAND                9
#define INVENTORY                  10
#define ITEM                       11
#define PROMOTION                  12
#define REASON                     13
#define SHIP_MODE                  14
#define STORE                      15
#define STORE_RETURNS              16
#define STORE_SALES                17
#define TIME                       18
#define WAREHOUSE                  19
#define WEB_PAGE                   20
#define WEB_RETURNS                21
#define WEB_SALES                  22
#define WEB_SITE                   23
#define DBGEN_VERSION              24
#define S_BRAND                    25
#define S_CUSTOMER_ADDRESS         26
#define S_CALL_CENTER              27
#define S_CATALOG                  28
#define S_CATALOG_ORDER            29
#define S_CATALOG_ORDER_LINEITEM   30
#define S_CATALOG_PAGE             31
#define S_CATALOG_PROMOTIONAL_ITEM 32
#define S_CATALOG_RETURNS          33
#define S_CATEGORY                 34
#define S_CLASS                    35
#define S_COMPANY                  36
#define S_CUSTOMER                 37
#define S_DIVISION                 38
#define S_INVENTORY                39
#define S_ITEM                     40
#define S_MANAGER                  41
#define S_MANUFACTURER             42
#define S_MARKET                   43
#define S_PRODUCT                  44
#define S_PROMOTION                45
#define S_PURCHASE                 46
#define S_PURCHASE_LINEITEM        47
#define S_REASON                   48
#define S_STORE                    49
#define S_STORE_PROMOTIONAL_ITEM   50
#define S_STORE_RETURNS            51
#define S_SUBCATEGORY              52
#define S_SUBCLASS                 53
#define S_WAREHOUSE                54
#define S_WEB_ORDER                55
#define S_WEB_ORDER_LINEITEM       56
#define S_WEB_PAGE                 57
#define S_WEB_PROMOTIONAL_ITEM     58
#define S_WEB_RETURNS              59
#define S_WEB_SITE                 60
#define S_ZIPG                     61
#define PSEUDO_TABLE_START         62
/* PSEUDO TABLES from here on; used in hierarchies */
#define ITEM_BRAND           62
#define ITEM_CLASS           63
#define ITEM_CATEGORY        64
#define DIVISIONS            65
#define COMPANY              66
#define CONCURRENT_WEB_SITES 67
#define ACTIVE_CITIES        68
#define ACTIVE_COUNTIES      69
#define ACTIVE_STATES        70
#define MAX_TABLE            70
#endif
