#include "src/storage/table/chunk_info.cpp"

#include "src/storage/table/column_checkpoint_state.cpp"

#include "src/storage/table/column_data_checkpointer.cpp"

#include "src/storage/table/column_data.cpp"

#include "src/storage/table/column_segment.cpp"

#include "src/storage/table/array_column_data.cpp"

#include "src/storage/table/list_column_data.cpp"

#include "src/storage/table/update_segment.cpp"

#include "src/storage/table/persistent_table_data.cpp"

#include "src/storage/table/row_group.cpp"

#include "src/storage/table/row_group_collection.cpp"

#include "src/storage/table/row_version_manager.cpp"

#include "src/storage/table/scan_state.cpp"

#include "src/storage/table/standard_column_data.cpp"

#include "src/storage/table/struct_column_data.cpp"

#include "src/storage/table/table_statistics.cpp"

#include "src/storage/table/validity_column_data.cpp"

