#include "src/planner/binder/statement/bind_attach.cpp"

#include "src/planner/binder/statement/bind_call.cpp"

#include "src/planner/binder/statement/bind_copy.cpp"

#include "src/planner/binder/statement/bind_copy_database.cpp"

#include "src/planner/binder/statement/bind_create.cpp"

#include "src/planner/binder/statement/bind_create_table.cpp"

#include "src/planner/binder/statement/bind_delete.cpp"

#include "src/planner/binder/statement/bind_detach.cpp"

#include "src/planner/binder/statement/bind_drop.cpp"

#include "src/planner/binder/statement/bind_execute.cpp"

#include "src/planner/binder/statement/bind_explain.cpp"

#include "src/planner/binder/statement/bind_export.cpp"

#include "src/planner/binder/statement/bind_extension.cpp"

#include "src/planner/binder/statement/bind_insert.cpp"

#include "src/planner/binder/statement/bind_load.cpp"

#include "src/planner/binder/statement/bind_logical_plan.cpp"

#include "src/planner/binder/statement/bind_pragma.cpp"

#include "src/planner/binder/statement/bind_prepare.cpp"

#include "src/planner/binder/statement/bind_relation.cpp"

#include "src/planner/binder/statement/bind_select.cpp"

#include "src/planner/binder/statement/bind_set.cpp"

#include "src/planner/binder/statement/bind_simple.cpp"

#include "src/planner/binder/statement/bind_summarize.cpp"

#include "src/planner/binder/statement/bind_update.cpp"

#include "src/planner/binder/statement/bind_vacuum.cpp"

