#include "src/catalog/catalog_entry.cpp"

#include "src/catalog/catalog.cpp"

#include "src/catalog/catalog_search_path.cpp"

#include "src/catalog/catalog_set.cpp"

#include "src/catalog/catalog_transaction.cpp"

#include "src/catalog/duck_catalog.cpp"

#include "src/catalog/dependency_manager.cpp"

#include "src/catalog/dependency_list.cpp"

#include "src/catalog/dependency_catalog_set.cpp"

#include "src/catalog/similar_catalog_entry.cpp"

